/**
 * Copyright 2008 Peach Jean Solutions
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.xj4;

import org.xj4.spi.MethodWrapperSource;
import org.xj4.spi.ClassWrapperSource;
import org.xj4.spi.TestWrapper;
import org.xj4.spi.MethodHandler;
import org.xj4.spi.MethodHandlerFactory;
import org.xj4.spi.XJ4ExtensionProvider;
import org.xj4.spi.TestClassValidator;

import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.io.IOException;
import java.lang.annotation.Annotation;

import org.junit.runner.notification.RunNotifier;
import org.junit.runners.BlockJUnit4ClassRunner;
import org.junit.runners.model.InitializationError;
import org.junit.runners.model.FrameworkMethod;
import org.junit.runners.model.Statement;
import org.junit.Test;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * This is a JUnit {@link org.junit.runner.Runner} that allows extending the basic functionality of JUnit.  Providers
 * that extend {@link XJ4ExtensionProvider} are provided the opportunity to hook into the test lifecycle at certain
 * places.  
 * @author jbunting
 *
 */
public class XJ4Runner extends BlockJUnit4ClassRunner {

  private static final Log logger = LogFactory.getLog(XJ4Runner.class);

  private XJ4ExtensionProvider extensionProvider;

  private boolean initialized = false;
  private List<MethodHandler> methodHandlers = new ArrayList<MethodHandler>();
  private XJ4TestClass testClass;

  /**
   * Creates a new runner for testing the provided {@code klass}.
   *
   * @param klass the class to test
   * @throws InitializationError if {@code klass} cannot be tested
   */
  public XJ4Runner(Class<?> klass) throws InitializationError {
    this(klass, null);
  }

  public XJ4Runner(Class<?> klass, XJ4ExtensionProvider extensionProvider) throws InitializationError {
    super(klass);
    this.testClass = new XJ4TestClass(klass);
    try {
      this.extensionProvider = extensionProvider != null ? extensionProvider : new ServiceExtensionProvider();
    } catch (IOException e) {
      throw new InitializationError(Arrays.<Throwable>asList(e));
    }
    for(MethodHandlerFactory factory: this.extensionProvider.getMethodHandlerFactories()) {
      methodHandlers.add(factory.create(this.getXJ4TestClass()));
    }
    initialized = true;
    validate();
  }

  @Override
  protected List<FrameworkMethod> computeTestMethods() {
    List<FrameworkMethod> original = super.computeTestMethods();
    List<FrameworkMethod> adjusted = new ArrayList<FrameworkMethod>();
    for(FrameworkMethod method: original) {
      boolean handled = false;
      for(MethodHandler handler: methodHandlers) {
        if(handler.canHandle(method)) {
          adjusted.addAll(handler.handle(method));
          handled = true;
          break;
        }
      }
      if(!handled) {
        adjusted.add(method);
      }
    }
    return adjusted;
  }

  @Override
  protected Statement withAfters(FrameworkMethod frameworkMethod, Object target, Statement statement) {
    final List<TestWrapper> wrappers = new ArrayList<TestWrapper>();
    for(MethodWrapperSource source: this.extensionProvider.getMethodWrapperSources()) {
      wrappers.addAll(source.generateWrapper(this.getXJ4TestClass(), frameworkMethod, target));
    }
    return new TestWrapperStatement(super.withAfters(frameworkMethod, target, statement), wrappers);
  }

  @Override
  protected Statement classBlock(RunNotifier runNotifier) {
    List<TestWrapper> wrappers = new ArrayList<TestWrapper>();
    for(ClassWrapperSource source: this.extensionProvider.getClassWrapperSources()) {
      wrappers.addAll(source.generateWrapper(this.getXJ4TestClass()));
    }
    return new TestWrapperStatement(super.classBlock(runNotifier), wrappers);
  }

  public XJ4TestClass getXJ4TestClass() {
    return this.testClass;
  }

  @Override
  protected void validateTestMethods(List<Throwable> errors) {
    validateMethods(Test.class, errors);
  }

  protected void validateMethods(Class<? extends Annotation> annotation, List<Throwable> errors) {
    List<FrameworkMethod> methods= getTestClass().getAnnotatedMethods(annotation);

    for (FrameworkMethod method : methods) {
      boolean handled = false;
      for(MethodHandler handler: methodHandlers) {
        if(handler.canHandle(method)) {
          handler.validate(method, errors);
          handled = true;
          break;
        }
      }
      if(!handled) {
        method.validatePublicVoidNoArg(false, errors);
      }
    }
  }

  protected void validate() throws InitializationError {
    List<Throwable> errors= new ArrayList<Throwable>();
    // calling the super method here, since in this class, we override it to do nothing
    collectInitializationErrors(errors);
    if (!errors.isEmpty())
      throw new InitializationError(errors);
  }

  @Override
  protected void collectInitializationErrors(List<Throwable> errors) {
    // if not initialized, do nothing, the parent calls this before we have a chance to setup our class,
    // causing all sorts of issues
    if(initialized) {
      super.collectInitializationErrors(errors);
      for(TestClassValidator validator : extensionProvider.getTestClassValidators()) {
        validator.validate(this.getXJ4TestClass(), errors);
      }
    }
  }
}
